//----------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                    (c) 1997-2010 European Space Agency
//
//---------------------------------------------------------------------------
//
//	SetTableColumn: --- a class representing a chain DataSet::Table::Column
//
//	DESCRIPTION:
//		Instances of class SetTableColumn represent a container for
//		a specification of a DataSet/Table/Column tuple. Table/Column
//		are optional. Syntax:
//
//			setname[:block[:column]] || setname::column || setname::
//
//	CHANGE HISTORY:
//		UL, 1999-04-14: 1.0.0: - first version
//		UL, 1999-04-27: 1.0.1: - bug fix in accessTable() (if a block
//								 name started with a digit this was interpreted
//								 as being a block specificaton by number)
//		UL, 1999-05-05: 1.1.0: - added accessMode argument to constructors
//		UL, 1999-11-17: 2.0.0: - SetArray -> SetBlock with methods to access
//								 block as Array or Table
//		UL, 2000-04-25: 2.1.0: - new static methods componentNames()
//
//	AUTHOR:
//		U. Lammers (UL), Wed Apr 14 17:54:08 MET DST 1999
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/selectlib/src/utils/SetTableColumn.cc,v 1.6 2009/06/26 15:51:59 aibarra Exp $
//

#include "stringutil.h"
#include "SetTableColumn.h"
#include "errstr.h"

#include <cassert>

SetTableColumn::SetTableColumn(const std::string &spec,
							   Dal::AccessMode accessMode,
							   DataSetServer::MemoryModel memModel)
	 : _accessMode(accessMode), _memModel(memModel),
		_set(0), _table(0), _column(0)
{
	std::vector<std::string> s(SetTableColumn::componentNames(spec));
	assert(s.size() == 3);
	_setName = s[0];
	_tableName = s[1];
	_columnName = s[2];
} // SetTableColumn::SetTableColumn()


std::vector<std::string> SetTableColumn::componentNames(
													const std::string &name)
{
	std::vector<std::string> v(3);

	std::string spec(name);
    std::string::size_type pos;
	std::string::size_type pos1 = 0;
	if ((pos = spec.rfind("+")) != std::string::npos) {
		//
		//	s+block
		//
		v[0] = spec.substr(0, pos);
		v[1] = spec.substr(pos+1, spec.length()-pos-1);
	} else if ((pos = spec.rfind("[")) != std::string::npos &&
			   (pos1 = spec.rfind("]")) != std::string::npos) {
		//
		//	s[block]
		//
		v[0] = spec.substr(0, pos);
		v[1] = spec.substr(pos+1, pos1-pos-1);
	} else {
		//
		//	s || s:: || s:: || s:block || s::column || s::block::column
		//
		std::vector<std::string> vec(breakup(spec, ":"));
		if (vec.size() > 2)
			v[2] = vec[2];
		if (vec.size() > 1) // AI gcc4 porting ambiguous 'else'
		{
			if (spec.rfind("::") != std::string::npos)
				v[2] = vec[1];
			else
				v[1] = vec[1];
		}
		if (vec.size() > 0)
			v[0] = vec[0];
	}

	//cout<<"|"<<v[0]<<"|"<<v[1]<<"|"<<v[2]<<"|\n";

	return v;
} // SetTableColumn::componentNames()


DataSet *SetTableColumn::accessSet() const
{
	if (!_set)
		_set = dataSetServer->open(_setName, _accessMode, _memModel);

	return _set;
} // SetTableColumn::accessSet()


DataSet *SetTableColumn::dataSet() const
{
	return accessSet();
} // SetTableColumn::dataSet()


Table *SetTableColumn::accessTable() const
{
	accessSet();
	if (!_table) // AI gcc4 porting ambiguous 'else'
	{
		if (!_tableName.length())
			//
			//	no table specifier given - access first block
			//
			_table = _set->table(0);
		else {
			//
			//	block specifier given - can be number or name
			//
			bool isNumber = true;
			for (unsigned n = 0; n<_tableName.length(); ++n)
				if (!isdigit(_tableName[n])) {
					isNumber = false;
					break;
				}
			if (isNumber) {
				//
				//	we have a number
				//
				int blockNr = toNumber<int>(_tableName) - 1;
				if (blockNr < 0)
					blockNr = 0;
				_table = _set->table(blockNr);
			} else
				//
				//	access table by name
				//
				_table = _set->table(_tableName);
		}
	}

	return _table;
}


Table *SetTableColumn::table() const
{
	return accessTable();
} // SetTableColumn::dataSet()


Column *SetTableColumn::column() const
{
	accessTable();
	if (!_column) // AI gcc4 porting ambiguous 'else'
	{
		if (!_columnName.length())
			//
			//	no column specifier given - access first one
			//
			_column = _table->column(0);
		else
			//
			//	column specifier given - can be number or name
			//
			if (isdigit(_columnName[0])) {
				//
				//	starts with digits, let's assume it is a number
				//
				int columnNr = toNumber<int>(_columnName);
				_column = _table->column(columnNr);
			} else
				//
				//	access table by name
				//
				_column = _table->column(_columnName);
	}

	return _column;
} // SetTableColumn::column()


SetTableColumn::~SetTableColumn()
{
} // ~SetTableColumn::SetTableColumn


std::ostream &operator<<(std::ostream &os, const SetTableColumn &tuple)
{
	os<<tuple.dataSet()->name()<<":"<<tuple.table()->name()<<":"
	  <<tuple.column()->name();

	return os;
} // operator<<



SetBlock::SetBlock(const std::string &spec, Dal::AccessMode accessMode,
				   DataSetServer::MemoryModel memModel)
	 : _accessMode(accessMode), _memModel(memModel), _set(0), _block(0)
{
	std::vector<std::string> s(SetBlock::componentNames(spec));
	assert(s.size() == 2);
	_setName = s[0];
	_blockName = s[1];
} // SetBlock::SetBlock()


std::vector<std::string> SetBlock::componentNames(const std::string &name)
{
	std::vector<std::string> v(2);

	std::string spec(name);
    std::string::size_type pos;
	std::string::size_type pos1 = 0;
	if ((pos = spec.rfind("+")) != std::string::npos ||
		(pos = spec.rfind(":")) != std::string::npos) {
		//
		//	s+block || s:block
		//
		v[0] = spec.substr(0, pos);
		v[1] = spec.substr(pos+1, spec.length()-pos-1);
	} else if ((pos = spec.rfind("[")) != std::string::npos &&
			   (pos1 = spec.rfind("]")) != std::string::npos) {
		//
		//	s[block]
		//
		v[0] = spec.substr(0, pos);
		v[1] = spec.substr(pos+1, pos1-pos-1);
	} else
		//
		//	s
		//
		v[0] = spec;

	// std::cout<<"|"<<v[0]<<"|"<<v[1]<<"|\n";

	return v;
} // SetBlock::componentName()


DataSet *SetBlock::accessSet() const
{
	if (!_set)
		_set = dataSetServer->open(_setName, _accessMode, _memModel);

	return _set;
} // SetBlock::accessSet()


DataSet *SetBlock::dataSet() const
{
	return accessSet();
} // SetBlock::dataSet()


Block *SetBlock::block() const
{
	accessSet();
	if (!_block) // AI gcc4 porting ambiguous 'else'
	{
		if (!_blockName.length())
			//
			//	no block specifier given - access first block
			//
			_block = _set->block(0);
		else
			//
			//	block specifier given - can be number or name
			//
			if (isdigit(_blockName[0])) {
				//
				//	starts with digits, let's assume it is a number
				//
				int blockNr = toNumber<int>(_blockName) - 1;
				if (blockNr < 0)
					blockNr = 0;
				_block = _set->block(blockNr);
			} else
				//
				//	access block by name
				//
				_block = _set->block(_blockName);
	}

	return _block;
} // SetBlock::block()

Table *SetBlock::table() const
{
	Block *bl = block();
	if (bl->type() != Block::TableT)
		errstr<<"Block `"<<bl->qualifiedName()<<"' is not a table"
			  <<Sas::error("NotATable");

	return dynamic_cast<Table *>(bl);
} // SetBlock::table()


Array *SetBlock::array() const
{
	Block *bl = block();
	if (bl->type() != Block::ArrayT)
		errstr<<"Block `"<<bl->qualifiedName()<<"' is not an array"
			  <<Sas::error("NotAnArray");

	return dynamic_cast<Array *>(bl);
} // SetBlock::array()


SetBlock::~SetBlock()
{
} // SetBlock::~SetBlock


std::ostream &operator<<(std::ostream &os, const SetBlock &pair)
{
	os<<pair.dataSet()->name()<<":"<<pair.block()->name();

	return os;
} // operator<<
